from rhme.config import Configuration
import numpy as np
import cv2
import os

config = Configuration()

number = np.load(config.package_path + '/tratamento/treated_data_three/not-number_testing_images.npz')
number = number['arr_0']
label_t = np.load(config.package_path + '/tratamento/treated_data_three/not-number_testing_labels.npz')
label_t = label_t['arr_0']
# label5 = label5['arr_0']

def exibir_imagem(image):
    for i in range(0, len(image), 100):
        cv2.imshow("Image", image[i])
        print(label_t[i])
        cv2.waitKey(0)

exibir_imagem(number)

cv2.destroyAllWindows()



from rhme.processamento_imagem import preprocessamento as preprocessamento
from rhme import helpers
import numpy as np
import cv2
import imutils
import os

class PosProcessamentoImagem:

    def __init__(self, image):
        self.image = image

    def segmentar_igualdade(self, symbols):
        xmin_sorted = sorted(symbols, key = lambda i: i['xmin'], reverse=False) 
        xmin_ymin_sorted = sorted(symbols, key = lambda i: (i['xmin'], i['ymin']), reverse=False) 

        try:
            c = 0
            i = 0
            candidate_index = 0

            for symbol in xmin_sorted:

                if c == 0 and symbol['label'] == '11':
                    # PRIMEIRO CANDIDATO

                    candidate = symbol.copy()
                    c = 1 # encontrou 1 candidato
                    candidate_index = i # armazena index do candidato

                elif c == 1 and symbol['label'] == '11':
                    # SEGUNDO CANDIDATO

                    '''
                    Fazer por %? A ordem importa.
                    x = candidate['xmin'] * 100 / symbol['xmin']
                    x = symbol['xmin'] * 100 / candidate['xmin']

                    Fazer pela diferença?
                    Vou conseguir definir um thresold?

                    Diferença x < média de tamanho dos traços.
                    '''

                    x = abs(candidate['xmin'] - symbol['xmin'])
                    mean = (candidate['w'] + symbol['w']) / 2

                    if x <= mean:
                        c = 0

                        abacate = self.image.copy()

                        ymin = min(symbol['ymin'], candidate['ymin'])
                        ymax = max(symbol['ymax'], candidate['ymax'])

                        xmin = min(symbol['xmin'], candidate['xmin'])
                        xmax = max(symbol['xmax'], candidate['xmax'])

                        cropped = abacate[
                            ymin:ymax + 1,
                            xmin:xmax + 1
                        ]

                        pre = preprocessamento.ProcessamentoImagem()
                        resized = pre.redimensionar(cropped)
                        resized = pre._255_to_1(resized)

                        symbol['image'] = resized
                        symbol['label'] = '10'
                        symbol['xmin'] = xmin
                        symbol['xmax'] = xmax
                        symbol['ymin'] = ymin
                        symbol['ymax'] = ymax
                        symbol['w'] = xmax - xmin
                        symbol['h'] = ymax - ymin
                        symbol['centroid'] = [xmin + (xmax-xmin) / 2, ymin + (ymax-ymin) / 2]

                        del xmin_sorted[candidate_index]

                    else:
                        candidate = symbol.copy()
                        candidate_index = i
                i += 1

        except BaseException as e:
            print(e)

        return xmin_sorted



from rhme.helpers import DataStructures as DS
from rhme.hme_parser import check_grammar as cg
from rhme import helpers
import re
import json
import numpy as np

helpers_labels = helpers.get_labels()
labels = helpers_labels['labels_parser']

class Parser:
    def __init__(self, expression):
        self.symbols = self.preprocessing(expression['symbols'])
        self.tree=None
        self.tlist=None

    def to_parse(self):
        try:
            tree = self.__main_parsing()
            self.tree=tree

            tlist = self.__tree_to_list(tree)
            self.tlist = tlist

            latex_before_cg = self.__list_to_latex_obj(tlist)

            check_grammar = cg.CheckGrammar()
            latex = check_grammar.check(latex_before_cg)
            lex_errors = check_grammar.lex_errors
            pure_lex_errors = check_grammar.pure_lex_errors
            yacc_errors = check_grammar.yacc_errors
            pure_yacc_errors = check_grammar.pure_yacc_errors


            return {
                'latex': latex, 
                'latex_before_cg': latex_before_cg, 
                'lex_errors': lex_errors, 
                'yacc_errors': yacc_errors,
                'pure_lex_errors': pure_lex_errors,
                'pure_yacc_errors': pure_yacc_errors
            }

        except Exception as e:
            raise e

    def __main_parsing(self):
        listin = self.symbols
        T = DS.Tree()
        Q = DS.Queue()
        S = DS.Stack()

        temp1 = 0
        temp2 = 0

        R = [[0,0], [9999999999, 9999999999]]
        sstart = self.__sp(listin, R)

        s = listin[sstart]

        R = [
            [s['wall']['left'], s['wall']['top']],
            [s['wall']['right'], s['wall']['bottom']]
        ]

        if sstart != -1:

            Q.enqueue(sstart)
            Q.enqueue(T.root_node)
            listin[sstart]['checked'] = True

        while not Q.is_empty():

            while not Q.is_empty():
                temp1 = Q.dequeue()
                ParentNode = Q.dequeue()
                SymbolNode = DS.SymbolNode(listin[temp1])
                T.insert(SymbolNode, ParentNode, 'Node')
                S.push(temp1)
                S.push(SymbolNode)

                temp2 = self.__hor(listin, temp1)

                while temp2 != -1:
                    listin[temp2]['checked'] = True
                    listin[temp2]['wall'] = listin[temp1]['wall'].copy()
                    SymbolNode = DS.SymbolNode(listin[temp2])
                    T.insert(SymbolNode, ParentNode, 'Node')
                    S.push(temp2)
                    S.push(SymbolNode)
                    listin[temp1]['wall']['right'] = listin[temp2]['xmin']
                    temp1 = temp2
                    temp2 = self.__hor(listin, temp1)
            S.push("EOBL")

            while not S.is_empty():
                if S.peek() == "EOBL":
                    S.pop()
                SymbolNode = S.pop()
                temp1 = S.pop()

                upperThreshold = listin[temp1]['ymin'] + ((1/6) * listin[temp1]['h'])
                lowerThreshold = listin[temp1]['ymin'] + ((5/6) * listin[temp1]['h'])

                R = [
                    {'above': [
                        [listin[temp1]['xmin'], listin[temp1]['wall']['top']],
                        [listin[temp1]['xmax'], upperThreshold]
                    ]},
                    {'below': [
                        [listin[temp1]['xmin'], lowerThreshold],
                        [listin[temp1]['xmax'], listin[temp1]['wall']['bottom']]
                    ]}
                ]

                for region in R:
                    reg = region[list(region.keys())[0]]
                    temp2 = self.__start(listin, reg)

                    if temp2 != -1:
                        if not listin[temp2]['checked']:
                            listin[temp2]['checked'] = True
                            listin[temp2]['wall']['left'] = reg[0][0]
                            listin[temp2]['wall']['right'] = reg[1][0]
                            listin[temp2]['wall']['top'] = reg[0][1]
                            listin[temp2]['wall']['bottom'] = reg[1][1]

                            RelationNode = DS.RegionNode(list(region.keys())[0])
                            T.insert(RelationNode, SymbolNode, 'Node')
                            Q.enqueue(temp2)
                            Q.enqueue(RelationNode)

                R = [
                    {'super': [
                        [listin[temp1]['xmax'], listin[temp1]['wall']['top']],
                        [listin[temp1]['wall']['right'], upperThreshold]
                    ]},
                    {'contains': [
                        [listin[temp1]['xmin'], listin[temp1]['ymin']],
                        [listin[temp1]['xmax'], listin[temp1]['ymax']]
                    ]},
                ]

                for region in R:
                    reg = region[list(region.keys())[0]]
                    temp2 = self.__start(listin, reg)

                    if temp2 != -1:
                        if not listin[temp2]['checked']:
                            listin[temp2]['checked'] = True
                            listin[temp2]['wall']['left'] = reg[0][0]
                            listin[temp2]['wall']['right'] = reg[1][0]
                            listin[temp2]['wall']['top'] = reg[0][1]
                            listin[temp2]['wall']['bottom'] = reg[1][1]

                            RelationNode = DS.RegionNode(list(region.keys())[0])
                            T.insert(RelationNode, SymbolNode, 'Node')
                            Q.enqueue(temp2)
                            Q.enqueue(RelationNode)

        return T

    def preprocessing(self, symbols):
        xmin_sorted = sorted(symbols, key = lambda i: i['xmin'])
        symbols = xmin_sorted

        for i in range(0,len(symbols)):
            s = symbols[i]
            s['centroid'] = list(s['centroid'])
            s['checked'] = False

            if s['label'] in ['12', '14', '16']:
                s['type'] = 'Open'
            elif s['label'] in ['13', '15', '17']:
                s['type'] = 'Close'
            else:
                s['type'] = 'Normal'

            if re.search("^[0-9]$", str(s['label']) or s['label'] == '20'):
                s['centroid_class'] = 'Ascending'
                s['centroid'][1] = s['ymin'] + (3/5) * (s['ymax'] - s['ymin'])
            elif s['label'] == 28 or s['type'] == 'Open' or s['label'] == '25':
                s['centroid_class'] = 'Descending'
                s['centroid'][1] = s['ymin'] + (1/4) * s['h']
            else:
                s['centroid_class'] = 'Centred'
                s['centroid'][1] = s['ymin'] + ((s['ymax'] - s['ymin'])/2)

            s['wall'] = {}
            s['wall']['top'] = -1
            s['wall']['bottom'] = 9999999999999
            s['wall']['left'] = -1
            s['wall']['right'] = 9999999999999

        return symbols

    def __overlap(self, symbolIndex, top, bottom, listin):
        listIndex = symbolIndex
        stop = False
        n = len(listin)

        if listin[symbolIndex]['label'] == '11':
            maxLength = listin[symbolIndex]['xmax'] - listin[symbolIndex]['xmin']
        else:
            maxLength = -1
        mainLine = -1

        while listIndex > 0 and stop == False:
            if listin[listIndex-1]['xmin'] < listin[symbolIndex]['xmin']:
                stop = True
            else:
                listIndex = listIndex - 1

        while listIndex < n and listin[listIndex]['xmin'] < listin[symbolIndex]['xmax']:

            if not listin[listIndex]['checked'] and \
                listin[listIndex]['label'] == '11' and \
                listin[listIndex]['centroid'][1] >= top and \
                listin[listIndex]['centroid'][1] < bottom and \
                listin[listIndex]['xmin'] <= (listin[symbolIndex]['xmin'] + 8) and \
                (listin[listIndex]['xmax'] - listin[listIndex]['xmin']) > maxLength:
                    maxLength = (listin[listIndex]['xmax'] - listin[listIndex]['xmin'])
                    mainLine = listIndex
            listIndex += 1

        if mainLine == -1:
            return symbolIndex
        else:
            return mainLine

    def __start(self, listin, R):
        left = R[0][0]
        top = R[0][1]
        right = R[1][0]
        bottom = R[1][1]

        leftmostIndex = -1
        listIndex = 0
        overlapIndex = -1
        n = len(listin)

        while leftmostIndex == -1 and listIndex < n:
            if not listin[listIndex]['checked'] and \
                listin[listIndex]['centroid'][0] > left and \
                listin[listIndex]['centroid'][1] > top and \
                listin[listIndex]['centroid'][0] < right and \
                listin[listIndex]['centroid'][1] < bottom:
                    leftmostIndex = listIndex
            else:
                listIndex = listIndex + 1

        if leftmostIndex == -1:
            return leftmostIndex
        else:
            return self.__overlap(leftmostIndex, top, bottom, listin)

    def __sp(self, listin, R):
        return self.__start(listin, R)

    def __hor(self, listin, index):
        if listin[index]['type'] == 'Open':
            left = listin[index]['wall']['left']
        else:
            left = listin[index]['xmax']

        right = listin[index]['wall']['right']

        if listin[index]['label'] == '11':
            top = listin[index]['wall']['top']
            bottom = listin[index]['wall']['bottom']
        else:
            top = listin[index]['ymin'] + (listin[index]['h'] * (1/6))
            bottom = listin[index]['ymin'] + (listin[index]['h'] * (5/6))

        stop = False

        c = 0
        global a

        for s in range(0, len(listin)):
            c += 1
            if not stop:
                checked = listin[s]['checked']
                if not checked:
                    symbol = listin[s]
                    if int(listin[index]['label']) in range(11,18):
                        R = [[left, top], [right, bottom]]
                        a = self.__start(listin, R)
                        stop = True
                    else: 
                        if symbol['centroid'][0] > left and \
                            symbol['centroid'][0] < right and \
                            symbol['centroid'][1] < bottom and \
                            symbol['centroid'][1] > top:
                                a = s
                                stop = True

        if stop and a != -1:
            return self.__overlap(a, listin[a]['wall']['top'], listin[a]['wall']['bottom'], listin)
        else:
            return -1

    def __tree_to_list(self, tree, node=None):

        latex = []

        def recur(root_node):
            current = tree.root_node if not root_node else root_node

            if current is None:
                return

            if isinstance(current.data, str):
                latex.append(current.data)
            else:
                try:
                    real_label = labels[current.data['label']]
                    if real_label == ",":
                        current.data['label'] = r"."
                    else:
                        current.data['label'] = real_label
                    latex.append(current.data)
                except BaseException as e:
                    print('Exception b: ', e)

            if current.node_type == 'RegionNode' or current.data['label'] == '25':
                latex.append('{')

            for node in current.children:
                recur(node)

            if current.node_type == 'RegionNode' or current.data['label'] == '25':
                latex.append('}')

        recur(node)

        if latex[0] == 'Expression':
            latex.remove('Expression')
            if latex[-1] == "}":
                latex.pop()
            if latex[0] == '{':
                latex.reverse()
                latex.pop()
                latex.reverse()

        return latex

    def __list_to_latex_obj(self, tlist):
        '''
        GARAIN, U.; CHAUDHURI, B. B. Recognition of Online Handwritten Mathematical Expressions. IEEE Transactions on Systems, Man, and Cybernetics, [s. l.], v. 34, n. 6, p. 2366–2376, 2004. 

        S -> ES | E
        E -> S^{S} | S_{S} | \\frac{S}{S} | \sqrt{S} | N | L | e 
        '''

        latex = []

        for symbol in tlist:
            if isinstance(symbol, dict):
                latex.append({
                    # 'label': int(symbol['label']) if symbol['label'] in numbers else symbol['label'],
                    'label': symbol['label'],
                    'prediction': symbol['prediction'] if 'prediction' in symbol else [],
                    'type': symbol['type'] or ''
                })
            else:
                latex.append({
                    'label': symbol,
                    'prediction': [],
                    'type': 'context'
                })

        grammar = {
            '-': 'frac',
            'below': 'below',
            'sqrt': 'sqrt',
            'super': 'super',
            '*': 'mult',
        }

        subst = {
            'frac': {
                '-': "\\frac",
                'above': '',
            },
            'below': {
                'below': ''
            },
            'sqrt': {
                'sqrt': '\\sqrt'
            },
            'super': {
                'super': '^'
            },
            'mult': {
                '*': '\cdot'
            }
        }

        for i in range(0, len(latex)):
            if latex[i]['label'] in grammar:
                subs = grammar[latex[i]['label']]
                for substitution in subst[subs]:
                    if latex[i]['label'] == substitution:
                        latex[i]['label'] = subst[subs][substitution]
                        i += 1

        return latex

if __name__ == "__main__":

    # for debug
    obj1 = {
        'symbols':
        [
            {'index': 2, 'xmin': 36, 'xmax': 102, 'ymin': 83, 'ymax': 161, 'w': 66, 'h': 78, 'centroid': [69.0, 80.5], 'label': '2'},
            {'index': 3, 'xmin': 109, 'xmax': 148, 'ymin': 48, 'ymax': 96, 'w': 39, 'h': 48, 'centroid': [128.5, 48.0], 'label': '2'},
            {'index': 0, 'xmin': 205, 'xmax': 245, 'ymin': 100, 'ymax': 147, 'w': 40, 'h': 47, 'centroid': [225.0, 73.5], 'label': '18'},
            {'index': 1, 'xmin': 300, 'xmax': 344, 'ymin': 88, 'ymax': 158, 'w': 44, 'h': 70, 'centroid': [322.0, 79.0], 'label': '5'}
        ]
    }

    obj2 = {
        'symbols':
        [
            {'index': 1, 'xmin': 143, 'xmax': 588, 'ymin': 197, 'ymax': 216, 'w': 445, 'h': 19, 'centroid': [365.5, 108.0], 'label': '11'},
            {'index': 4, 'xmin': 174, 'xmax': 534, 'ymin': 32, 'ymax': 199, 'w': 360, 'h': 167, 'centroid': [354.0, 99.5], 'label': '25'}, 
            {'index': 2, 'xmin': 280, 'xmax': 364, 'ymin': 92, 'ymax': 173, 'w': 84, 'h': 81, 'centroid': [322.0, 86.5], 'label': '19'},
            {'index': 0, 'xmin': 302, 'xmax': 428, 'ymin': 254, 'ymax': 358, 'w': 126, 'h': 104, 'centroid': [365.0, 179.0], 'label': '27'}, 
            {'index': 3, 'xmin': 379, 'xmax': 477, 'ymin': 60, 'ymax': 166, 'w': 98, 'h': 106, 'centroid': [428.0, 83.0], 'label': '20'}
        ]
    }

    a = Parser(obj1)
    print(a.to_parse())



from rhme.hme_parser.grammar import yacc as yacc
from rhme.hme_parser.grammar import lex as lex
from rhme.hme_parser import base_grammar as base_grammar
from rhme.hme_parser import check_grammar_lex as check_grammar_lex
from rhme.helpers.exceptions import GrammarError, LexicalError, SintaticError
from rhme import helpers
import numpy as np

class CheckSintax():
    def __init__(self):
        self.__first_error = True
        self.latex_string=""
        self.latex=""
        self.latex_list=""
        self.attempts=0
        self.index=0
        self.token_errors_history=[]
        self.yacc_error_list=None
        self.symbols_attempts_in_lex=[]
        self.grammar_errors_history=[]
        self.pure_yacc_errors=[]

    def __locate_grammar_error(self):
        helpers.debug("[check_grammar_sintax.py] __locate_grammar_error() | Locating all errors and creating a data structure.")

        yacc_error_list = self.yacc_error_list.copy()
        latex = self.latex.copy()
        grammar_errors = []
        grammar_errors_history = []
        
        for error in yacc_error_list:

            if error['value'] != None:
                latex_error = self.latex_string[error['lexpos']::]
                latex_error_token = error['value']
                latex_error_pos = error['lexpos']

                count = 0
                count_list = 0

                for symbol in latex:
                    if symbol['label'] != latex_error_token:
                        count += len(symbol['label'])
                        count_list += 1
                    elif count == latex_error_pos:
                        grammar_errors.append({
                            'pos': latex_error_pos,
                            'pos_list': count_list,
                            'label': symbol['label'],
                            'prediction': symbol['prediction'],
                            'attempts': [latex_error_pos]
                        })
                        grammar_errors_history.extend(grammar_errors)
                        break
            else:
                helpers.debug("Use automata to fix")
                continue

        return grammar_errors, grammar_errors_history

    def check_correct_grammar(self):
        helpers.debug("[check_grammar_sintax.py] check_correct_grammar()")
        helpers.debug("[check_grammar_lex.py] check_correct_grammar() | attempts: %s" % self.attempts)

        second_yacc_error_list = None
        grammar_errors = []

        if not self.yacc_error_list and self.__first_error and self.attempts < 3:
            yacc_error_list = yacc.LatexParse(self.latex_string)
            print('YACC ERROR: ', yacc_error_list)
            self.pure_yacc_errors.append(yacc_error_list)
            self.yacc_error_list = yacc_error_list

            if self.yacc_error_list:

                grammar_errors, grammar_errors_history = self.__locate_grammar_error()
                self.grammar_errors_history.extend(grammar_errors_history)

                self.__first_error = False

                # TESTE
                # return self.latex_string, grammar_errors, self.grammar_errors_history, self.latex, self.latex_list
                return self.__attempt_to_fix_error(grammar_errors)

                # self.__attempt_to_fix_error(grammar_errors)

                raise GrammarError({
                    'error': grammar_errors,
                    'latex': self.latex,
                    'latex_list': self.latex_list,
                    'grammar_errors_history': self.grammar_errors_history,
                    'latex_string': self.latex_string,
                    'pure_yacc_errors_list': self.pure_yacc_errors
                })

        elif self.yacc_error_list and not self.__first_error and self.attempts < 3:

            self.__find_lexical_errors()

            second_yacc_error_list = yacc.LatexParse(self.latex_string)
            self.pure_yacc_errors.append(second_yacc_error_list)

            if second_yacc_error_list:

                second_yacc_error_list = self.__get_only_current_error(second_yacc_error_list)
                self.yacc_error_list = second_yacc_error_list
                
                grammar_errors, grammar_errors_history = self.__locate_grammar_error()

                self.__attempt_to_fix_error(grammar_errors)

                raise GrammarError({
                    'error': grammar_errors,
                    'latex': self.latex,
                    'latex_list': self.latex_list,
                    'grammar_errors_history': self.grammar_errors_history,
                    'latex_string': self.latex_string,
                    'pure_yacc_errors_list': self.pure_yacc_errors
                })

        elif self.yacc_error_list and attempts >= 3:
            raise GrammarError({
                    'error': grammar_errors,
                    'latex': self.latex,
                    'latex_list': self.latex_list,
                    'grammar_errors_history': self.grammar_errors_history,
                    'latex_string': self.latex_string,
                    'pure_yacc_errors_list': self.pure_yacc_errors
                })

        return self.latex_string, grammar_errors, self.grammar_errors_history, self.latex, self.latex_list

    def __find_lexical_errors(self):
        cgl = check_grammar_lex.CheckLex()
        cgl.latex_string = self.latex_string
        cgl.latex = self.latex
        cgl.latex_list = self.latex_list
        # cgl.attempts = self.attempts-1
        cgl.attempts = 0

        new_latex_string, token_errors, token_errors_history, latex, latex_list = cgl.check_correct_lex()

        self.latex_string = new_latex_string
        self.token_errors_history = token_errors_history
        self.latex = latex
        self.latex_list = latex_list

        if token_errors:
            for token in token_errors_history:
                self.symbols_attempts_in_lex.extend(token['attempts'])

    def __get_only_current_error(self, second_yacc_error_list):
        """Se erro novo for EOF ou se for em uma posição diferente do PRIMEIRO símbolo com erro')
        Remove erro da lista. Para resolver apenas o erro atual / primeiro erro | usa sempre index=0
        """
        cond = False
        if isinstance(second_yacc_error_list[self.index], dict):
            helpers.debug('Sintatic error')
            print(second_yacc_error_list)
            cond = second_yacc_error_list[self.index]['lexpos'] != None and second_yacc_error_list[0]['lexpos'] != self.yacc_error_list[0]['lexpos']
        if isinstance(second_yacc_error_list[self.index], tuple):
            helpers.debug('Lex error')
            cond = second_yacc_error_list[self.index][1] != -1 and second_yacc_error_list[self.index][1] != self.yacc_error_list[self.index]['lexpos']
        if cond:
            # Remove último erro adicionado para manter no anterior
            second_yacc_error_list.reverse()
            second_yacc_error_list.pop()
            second_yacc_error_list.reverse()
        return second_yacc_error_list

    def __attempt_to_fix_error(self, grammar_errors):
        bg = base_grammar.BaseGrammar()
        fix_attempts = self.symbols_attempts_in_lex.copy()
        fix_attempts.extend(self.grammar_errors_history)
        updated_latex_string, updated_grammar_error_list, self.index = bg.correct_grammar_lex(grammar_errors, self.latex, self.latex_list, self.index, fix_attempts)

        self.yacc_error_list = updated_grammar_error_list
        if updated_grammar_error_list:
            self.grammar_errors_history.extend(updated_grammar_error_list)

        if updated_latex_string:
            self.latex_string = updated_latex_string
            self.attempts += 1
            return self.check_correct_grammar()



from rhme import helpers
from rhme.hme_parser import check_grammar_lex as check_grammar_lex
from rhme.hme_parser import check_grammar_sintax as check_grammar_sintax
from rhme.helpers.exceptions import GrammarError, LexicalError, SintaticError

class CheckGrammar():

    def __init__(self):
        self.__attempts_lex = 0
        self.__attempts_grammar = 0
        self.lex_errors=None
        self.pure_lex_errors=None
        self.yacc_errors=None
        self.pure_yacc_errors=None

    def __remove_contains(self, tstring):
        if isinstance(tstring, str):
            tstring = tstring.replace('contains', '')
        elif isinstance(tstring, list) and "contains" in tstring:
            tstring = tstring.remove('contains')
        return tstring

    def check(self, tstring):
        helpers.debug("[check_grammar.py] check()")

        latex = []
        latex_list = []

        print('check() antes do for')
        # Revisar
        for symbol in tstring:
            if symbol['label'] != '' and symbol['label'] != 'contains':
                latex.append(symbol)
                latex_list.append(symbol['label'])

        print('check() depois do for')

        latex_string = "".join(latex_list)
        print('depois latex string')
        latex_string = self.__remove_contains(latex_string)
        print('depois latex string')
        lstring = latex_string

        helpers.debug("Latex List:")
        helpers.debug(latex_list)
        helpers.debug("Latex String: %s " % latex_string)

        '''
        Latex: {'label': symbol, 'prediction': [], 'type': 'context'}
        Latex List: ['\sqrt','{','9','}']
        Latex String: '\sqrt{9}'
        '''

        try:
            new_latex_string, token_errors, token_errors_history, latex, latex_list = self.__check_lex(latex_string, latex, latex_list)
            print('check_grammar.py    after check lex')
            self.lex_errors = token_errors_history

            print("=======================================")
            # teste
            if not token_errors:
                new_latex_string, grammar_errors, grammar_errors_history, latex, latex_list = self.__check_yacc(token_errors_history, new_latex_string, latex, latex_list)
                self.yacc_errors = grammar_errors_history

            else:
                raise LexicalError(data={
                    'error': token_errors,
                    'latex': latex,
                    'latex_list': latex_list,
                    'token_errors_history': token_errors_history,
                    'latex_string': latex_string,
                    'new_latex_string': new_latex_string
                })

            if new_latex_string != -1 and new_latex_string != None:
                lstring = new_latex_string

            return lstring
        except Exception as e:
            print('check() e: ', e)
            raise e

    def __check_lex(self, latex_string, latex, latex_list):
        cgl = check_grammar_lex.CheckLex()
        cgl.latex_string = latex_string
        cgl.latex = latex
        cgl.latex_list = latex_list
        cgl.attempts = self.__attempts_lex
        new_latex_string, token_errors, token_errors_history, latex, latex_list = cgl.check_correct_lex()
        self.pure_lex_errors = cgl.pure_lex_errors
        return new_latex_string, token_errors, token_errors_history, latex, latex_list

    def __check_yacc(self, token_errors_history, new_latex_string, latex, latex_list):
        symbols_attempts_in_lex = []
        for token in token_errors_history:
            symbols_attempts_in_lex.extend(token['attempts'])

        cgs = check_grammar_sintax.CheckSintax()
        cgs.latex_string = new_latex_string
        cgs.latex = latex
        cgs.latex_list = latex_list
        cgs.attempts = self.__attempts_grammar
        cgs.symbols_attempts_in_lex = symbols_attempts_in_lex                
        result = cgs.check_correct_grammar()
        self.pure_yacc_errors = cgs.pure_yacc_errors
        return result


from rhme import helpers
import numpy as np
import cv2
import imutils
import os

class ProcessamentoImagem:

    def __init__(self, configs={}):
        self.configs = {
            'black': False,
            'dilate': False,
            'dataset': False,
            'erode': False,
            'resize': 'smaller'
        }

        if configs:
            self.configs.update(configs)

    def print_bounding_box(self, image, coordinates):
        image_rect = image
        x, y, w, h = coordinates
        cv2.rectangle(image_rect,(x,y),(x+w,y+h),(100,100,100),2)
        # helpers.exibir_imagem(image_rect)
        return image_rect
        
    def save_image_255(self, image, name):
        img = cv2.convertScaleAbs(image, alpha=(255.0))
        cv2.imwrite('%s.jpg' % name, img)

    def to_gray_denoise(self, image):
        img = image.copy()

        if img.ndim == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        img = cv2.fastNlMeansDenoising(img, None, 5, 9)
        img = np.array(img)

        return img

    def invert(self, image):
        img = image.copy()
        return 255-img

    def binarization(self, image):
        img = image.copy()
        img = self.invert(img)
        img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)
        img = self.invert(img)
        return img

    def normalizar(self, image):
        img = image.copy()
        img = self.to_gray_denoise(image)

        if not self.configs['black']:
            img = self.invert(img)

        kernel = np.ones((2,2),np.uint8)

        if 'dilate' in self.configs:
            if self.configs['dilate']:
                img = cv2.dilate(img, kernel, iterations = 2)

        if 'erode' in self.configs:
            if self.configs['erode']:
                img = cv2.erode(img, kernel, iterations = 1)

        return img

    def _255_to_1(self, image):
        img = image.copy()
        return (img / 255)

    def redimensionar(self, image):
        old_size = image.shape[:2] # (height, width)
        height, width = old_size[0], old_size[1]

        ratio = float(26) / max(old_size)
        size = tuple([int(x*ratio) for x in old_size])

        size_height, size_width = size[0], size[1]

        division_height = int(height/2)
        division_width = int(width/2)

        middle_height = [
            image[division_height-1][division_width],
            image[division_height][division_width],
            image[division_height+1][division_width]
        ]
        middle_width = [
            image[division_height][division_width-1],
            image[division_height][division_width],
            image[division_height][division_width+1]
        ]

        if size_height <= 14 and size_width >= 20 and \
            (any(i > 0.0000 for i in middle_height) or any(i > 0.0000 for i in middle_width)):
            # For horizontal line
            nsize = 5 if size_height < 5 else size_height
            new_size = tuple([int(nsize), 26])
        else:
            if size_width / size_height >= 2:
                # For rectangle (sqrt)
                print('ractangle')
                kernel = np.ones((2,2),np.uint8)
                image = cv2.dilate(image, kernel, iterations = 20)
                xinit = int(width * 2 / 100)
                xend = int(width * 65 / 100)
                image = image[0:height, xinit:xend]

            new_size = size

        if self.configs['resize'] == 'smaller':
            image = cv2.resize(image.copy(), (new_size[1], new_size[0]), interpolation=cv2.INTER_AREA)
        elif self.configs['resize'] == 'bigger':
            image = cv2.resize(image.copy(), (new_size[1], new_size[0]), interpolation=cv2.INTER_LINEAR)

        # Cria borda ao redor do símbolo e normaliza para 28x28 px
        delta_w = 28 - new_size[1]
        delta_h = 28 - new_size[0]
        top, bottom = delta_h//2, delta_h-(delta_h//2)
        left, right = delta_w//2, delta_w-(delta_w//2)
        color = [0, 0, 0]
        image = cv2.copyMakeBorder(image.copy(), top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)
        return image

    def resize_full_image(self, image):
        h, w = image.shape[:2]
        if w > 4000:
            width = w * 20 / 100
            r = width / float(w)
            size = (width, int(h * r))
            image = cv2.resize(image, size)
        return image

    def segmentar(self, img):
        image = img.copy()
        symbols = []
        cnts, somethingElse = cv2.findContours(image.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        for i in range(len(cnts)):
            if(cv2.contourArea(cnts[i]) < 0):
                continue

            if(self.configs['dataset'] and len(cnts) > 1):
                continue

            try:
                # Draw contour in new image (mask)
                mask = np.zeros_like(image)
                cv2.drawContours(mask, cnts, i, (255, 255, 255), -50)
                out = np.zeros_like(image)

                # nas posições onde a máscara é 255, pinta as posições de normalized com as mesmas de mask 255
                # mask tem a parte interna do "2" pintada, mas o normalized não tem
                # mudei para > 0 em vez de 255 - isso evita que a imagem precise ser binarizada
                out[mask > 0] = image[mask > 0]

                # # Get bounding box coordinates
                _x, _y, _w, _h = cv2.boundingRect(cnts[i])

                # Por enquanto, não serve pra nada. Já tenho a informação ali em cima
                list_y, list_x = np.where(out > 0) #Diz TODOS os pontos onde a máscara == 255 | > 0
                (topx, topy) = (np.min(list_x), np.min(list_y))
                (bottomx, bottomy) = (np.max(list_x), np.max(list_y))

                # Crop image
                ycrop = _y + _h + 1
                xcrop = _x + _w + 1
                cropped = out[_y: ycrop, _x: xcrop]

                resized = self.redimensionar(cropped)
                result_image = self._255_to_1(resized)

                attributes = {
                    'index': i,
                    'image': result_image.copy(),
                    'xmin': _x,
                    'xmax': _x+_w,
                    'ymin': _y,
                    'ymax': _y+_h,
                    'w': _w,
                    'h': _h,
                    'centroid': [(_x + (_x+_w))/2, (_y + (_y+_h))/2]
                }

                symbols.append(attributes)

                mask = None
                out = None
                cropped = None
                resized = None
                binarized = None
                result_image = None

                self.image = self.print_bounding_box(image, (_x, _y, _w, _h))

            except BaseException as e:
                print(e)
                continue

        return (symbols, self.image)

    def mnist(self, img):
        try:
            if type(img) is str:
                image = cv2.imread(img)
            else:
                image = img

            normalizard = self.normalizar(image)
            result_image = self._255_to_1(normalizard)
            return result_image
        except BaseException as e:
            print(e)
            return None

    def tratamento_sem_segmentar(self, img):
        try:
            if type(img) is str:
                image = cv2.imread(img)
            else:
                image = img

            normalizard = self.normalizar(image)
            resized = self.redimensionar(normalizard)
            result_image = self._255_to_1(resized)
            return result_image
        except BaseException as e:
            print(e)
            return None

    def tratamento(self, img):
        try:
            if type(img) is str:
                image = cv2.imread(img)
            else:
                image = img

            original = image.copy()
            image = self.resize_full_image(image)
            normalizard = self.normalizar(image)
            self.image = normalizard.copy()

            symbols = self.segmentar(normalizard)

            return symbols
        except BaseException as e:
            print(e)
            return []

if __name__ == "__main__":

    p = ProcessamentoImagem()
    segmentacao, image = p.tratamento('images/numbers/teste10.png')

    try:
        for s in segmentacao:
            s.__delitem__('image')
    except Exception as identifier:
        print(identifier)

    print(segmentacao)



from rhme.hme_parser.grammar import yacc as yacc
from rhme.hme_parser.grammar import lex as lex
from rhme.hme_parser import base_grammar as base_grammar
from rhme import helpers
from rhme.helpers.exceptions import GrammarError, LexicalError, SintaticError
import numpy as np

class CheckLex():
    def __init__(self):
        self.__first_error = True
        self.latex_string=""
        self.latex=""
        self.latex_list=""
        self.attempts=0
        self.lex_error_list=None
        self.index=0
        self.token_errors_history=[]
        self.pure_lex_errors=[]

    def __locate_lex_error(self):
        helpers.debug("[check_grammar_lex.py] __locate_lex_error() | Locating all errors and creating a data structure.")

        error_list = self.lex_error_list.copy()
        latex = self.latex.copy()
        token_errors = []

        helpers.debug("[check_grammar_lex.py] __locate_lex_error() | List of lex errors:")
        helpers.debug(error_list)

        for error in error_list:

            if isinstance(error, dict):
                error = (error['label'], error['pos'])

            if error[1] != -1:
                helpers.debug("[check_grammar_lex.py] __locate_lex_error() | Locating error in latex_string.")
                latex_error_pos = error[1]
                latex_error = self.latex_string[latex_error_pos::]
                latex_error_token = error[0]

                count = 0
                count_list = 0

                helpers.debug("[check_grammar_lex.py] __locate_lex_error() | Searching token (in latex) with position returned in error.")
                for symbol in latex:

                    if symbol['label'] != latex_error_token:
                        # Increment counter if current position is the position containing the error
                        count += len(symbol['label'])
                        count_list += 1
                    elif count == latex_error_pos:
                    # elif count == latex_error_pos and \
                    # (symbol['label'] == "{" or symbol['label'] == "}" or
                    # symbol['label'] == "(" or symbol['label'] == ")" or
                    # symbol['label'] == "[" or symbol['label'] == "]"):
                        # If found the token at the error position and token is one of those symbols, append error.
                        token_errors.append({
                            'pos': latex_error_pos,
                            'pos_list': count_list,
                            'label': symbol['label'],
                            'prediction': symbol['prediction'],
                            'attempts': [latex_error_pos] # Stores the symbol attempts to restore the grammar
                        })
                        self.token_errors_history.extend(token_errors)
                        break
            else:
                helpers.debug("[check_grammar_lex.py] __locate_lex_error() | Cannot fix the error. Use an automata do fix it. Logging the error:")
                helpers.debug(error)
                continue

        return token_errors, self.token_errors_history

    def check_correct_lex(self):
        """Check and correct lex errors

        Args:
            latex_string (str): Latex string.
            latex (list): First latex structure.
            latex_list (list): [description]

        Returns:
            latex_string (str):
            token_errors (list): 
            token_errors_history (list):
            latex (list):
            latex_list (list):
        """

        helpers.debug("[check_grammar_lex.py] check_correct_lex()")
        helpers.debug("[check_grammar_lex.py] check_correct_lex() | attempts: %s" % self.attempts)
        helpers.debug("[check_grammar_lex.py] check_correct_lex() | error list:")
        helpers.debug(self.lex_error_list)
        helpers.debug(self.__first_error)

        second_lex_error_list = None
        token_errors = []
        print('hein?')
        if not self.lex_error_list and self.__first_error and self.attempts < 3:
            print('first')
            helpers.debug("[check_grammar_lex.py] check_correct_lex() | There's no previous error. Searching the first one.")

            lex_error_list = lex.LatexLexer(self.latex_string)
            print('LEX ERROR: ', lex_error_list)

            # teste
            self.pure_lex_errors.append(lex_error_list)
            self.lex_error_list = lex_error_list

            if self.lex_error_list:
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | First error found in lex.")
                token_errors, self.token_errors_history = self.__locate_lex_error()

                # Recursion to fix the error.
                print('first error: attempt....')
                print('latex list: ', self.latex_list)
                self.__first_error = False
                self.__attempt_to_fix_error(token_errors)

                if token_errors:
                    raise LexicalError({
                        'error': token_errors,
                        'latex': self.latex,
                        'latex_list': self.latex_list,
                        'token_errors_history': self.token_errors_history,
                        'latex_string': self.latex_string,
                        'pure_lex_errors_list': self.pure_lex_errors
                    })

        elif self.lex_error_list and not self.__first_error and self.attempts < 3:
            print('middle')
            helpers.debug("[check_grammar_lex.py] check_correct_lex() | There's previous error. Searching for new errors.")

            second_lex_error_list = lex.LatexLexer(self.latex_string)
            self.pure_lex_errors.append(second_lex_error_list)

            if second_lex_error_list:
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | New errors found.")

                ''' 
                    If new error is EOF error or, if it is at a different position from the first symbol with error
                    Remove the error from the list. Takes the next one.
                    One step at the time. One error at the time.
                '''
                if second_lex_error_list[0][1] != -1 and second_lex_error_list[0][1] != self.lex_error_list[0]['pos']:
                    self.lex_error_list.reverse()
                    self.lex_error_list.pop()
                    self.lex_error_list.reverse()

                token_errors = self.lex_error_list

               # Recursion to fix the error.
                print('more error: attempt....')
                print('latex list: ', self.latex_list)
                self.__attempt_to_fix_error(token_errors)

                raise LexicalError({
                    'error': token_errors,
                    'latex': self.latex,
                    'latex_list': self.latex_list,
                    'token_errors_history': self.token_errors_history,
                    'latex_string': self.latex_string,
                    'pure_lex_errors_list': self.pure_lex_errors
                })

        elif self.lex_error_list and self.attempts >= 3:
            print('last')
            raise LexicalError({
                'error': token_errors,
                'latex': self.latex,
                'latex_list': self.latex_list,
                'token_errors_history': self.token_errors_history,
                'latex_string': self.latex_string,
                'pure_lex_errors_list': self.pure_lex_errors
            })

        print('no more errors?')
        return self.latex_string, token_errors, self.token_errors_history, self.latex, self.latex_list

    def __attempt_to_fix_error(self, token_errors):
        helpers.debug("[check_grammar_lex.py] self.__attempt_to_fix_error() | Tries to fix the error.")
        bg = base_grammar.BaseGrammar()
        update_latex_string, updated_lex_error_list, self.index = bg.correct_grammar_lex(token_errors, self.latex, self.latex_list, self.index)

        helpers.debug("[check_grammar_lex.py] self.__attempt_to_fix_error() | Save errors in history.")
        self.lex_error_list = updated_lex_error_list
        if updated_lex_error_list:
            self.token_errors_history = updated_lex_error_list

        if update_latex_string:
            self.latex_string = update_latex_string
            self.attempts += 1
            helpers.debug("[check_grammar_lex.py] self.__attempt_to_fix_error() | Recursion. Check tokens again.")
            return self.check_correct_lex()


from rhme.config import Configuration
import os
import cv2 as cv
import idx2numpy
import json

config = Configuration()

def criar_diretorio(name):
    dirname = name
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    return dirname

def exibir_imagem(image):
    conf = config.get_configs()
    if conf["application"]["debug_mode_image"] == "active":
        cv.namedWindow("Imagem", cv.WINDOW_NORMAL)
        # cv.resizeWindow("Imagem", 725, 325)
        # cv.moveWindow("Imagem", 40,525)
        cv.imshow("Imagem", image)
        cv.waitKey(0)
        cv.destroyAllWindows()

def salvar_imagem(self, dirname, roi, i):
        filename = "char" + str(i) + ".png"
        cv2.imwrite(dirname+ "/" + filename, roi)

def get_labels():
    try:
        with open(config.package_path + '/docs/config/config_all.json') as json_file:
            labels_json = json_file.read()
            labels_dict = json.loads(labels_json)
            labels = labels_dict
    except Exception as e:
        print(e)
        labels = {}

    return labels

def save_json_file(path, data):
    with open(path, 'w') as f:
        f.write(json.dumps(data))

def debug(data):
    conf = config.get_configs()
    if conf["application"]["debug_mode"] == "active":
        print(data)






import json

class GrammarError(Exception):
    def __init__(self, valor="Could not solve the grammar error", data=None):
        self.valor = valor
        self.data = data
    def __repr__(self):
        return {
            'message': self.valor,
            'data': self.data
        }
    def __str__(self):
        return self.__repr__()

class SintaticError(Exception):
    def __init__(self, valor="Could not solve the sintatic error", data=None):
        self.valor = valor
        self.data = data
    def __repr__(self):
        return {
            'message': self.valor,
            'data': self.data
        }
    def __str__(self):
        return self.__repr__()

class LexicalError(Exception):
    def __init__(self, valor="Could not solve the lexical error", data=None):
        self.valor = valor
        self.data = data
    def __repr__(self):
        return {
            'message': self.valor,
            'data': self.data
        }
    def __str__(self):
        return self.__repr__()



from rhme.api import *
from rhme.config import Configuration
import base64
import numpy as np
import cv2

configs = Configuration()

class Example:

    def __init__(self):
        hme_recognizer = HME_Recognizer()

        # images = [configs.package_path + '/images/numbers/teste3.png', configs.package_path + '/images/numbers/teste4.png', configs.package_path + '/images/numbers/teste8.png', configs.package_path + '/images/numbers/teste10.png', configs.package_path + '/images/numbers/teste11.png']
        images = [
            configs.package_path + '/images/validacao/1.png', 
            configs.package_path + '/images/validacao/2.png', 
            configs.package_path + '/images/validacao/3.png'
        ]

        # with open(config.package_path + '/image.json') as json_file:
        #     labels_json = json_file.read()
        #     data = json.loads(labels_json)
        # img_data = data['image']
        # img_data = img_data.split(',')[1]
        # im_bytes = base64.b64decode(img_data)
        # im_arr = np.frombuffer(im_bytes, dtype=np.uint8) 
        # img = cv2.imdecode(im_arr, flags=cv2.IMREAD_COLOR)
    
        for image in images:
            try:
                expression, img = hme_recognizer.recognize(image)
                print("\nExpressão: ", expression)
            except Exception as e:
                print('Exception: ', e)

Example()



from rhme.reconhecimento import reconhecimento as reconhecimento
from rhme.processamento_imagem import preprocessamento as preprocessamento
from rhme.processamento_imagem import posprocessamento as posprocessamento
from rhme import helpers
import os
import re
import json
import numpy as np
import json
import cv2 as cv

helpers_labels = helpers.get_labels()
labels = helpers_labels['labels_parser']

class Recognize:

    def __init__(self, image):
        self.prediction=[]
        self.image = None

        try:
            if isinstance(image,str):
                nparr = np.fromstring(image, np.uint8)
                img = cv.imdecode(nparr, cv.IMREAD_COLOR)
            else:
                img = image.copy()
            self.image = img
        except Exception as e:
            print("[recognize.py] __init__ | Exception:")
            raise e

    def __recognize(self, img):
        img = img.reshape(1, 28, 28, 1)
        return reconhecimento.fit(img)

    def to_recognize(self):

        expression = {}

        helpers.debug("[recognize.py] to_recognize | Exibindo imagem da expressão...\n")
        helpers.exibir_imagem(self.image)

        helpers.debug("[recognize.py] to_recognize | Iniciando pré-processamento de imagem...\n")
        p = preprocessamento.ProcessamentoImagem()
        segmentacao, normalized_image = p.tratamento(self.image)

        helpers.debug("[recognize.py] to_recognize | Exibindo imagem pré-processada\n")
        helpers.exibir_imagem(normalized_image)

        helpers.debug("[recognize.py] to_recognize | Pré-processamento de imagem finalizado.\n")

        helpers.debug("[recognize.py] to_recognize | Iniciando classificação dos símbolos...\n")
        try:
            for s in segmentacao:
                helpers.exibir_imagem(s['image'])
                reconhecer = self.__recognize(s['image'])
                reconhecer['label'] = str(reconhecer['label'])

                s['label'] = reconhecer['label']
                s['prediction'] = reconhecer['prediction']        

                symbol_prediction = {
                    'identity': labels[s['label']]
                }
                symbol_prediction.update(reconhecer)
                self.prediction.append(symbol_prediction)

                helpers.debug('[recognize.py] to_recognize | Símbolo: %s ' % labels[s['label']])
            helpers.debug("\n[recognize.py] to_recognize | << Classificação dos símbolos finalizada.\n")
        except Exception as e:
            print("[recognize.py] to_recognize | Exception:")
            raise e
            

        pos = posprocessamento.PosProcessamentoImagem(normalized_image)
        new = pos.segmentar_igualdade(segmentacao)

        expression['symbols'] = new

        return (expression, normalized_image)



import cv2 as cv
from rhme import helpers
from rhme.hme_parser import parser as parser
from rhme.recognize import *
from rhme.config import Configuration as config

class HME_Recognizer:

    def __init__(self):
        self.reset()

    def __to_parse(self, expression):
        try:
            parse = parser.Parser(expression)
            to_parse = parse.to_parse()

            parsed_expression = to_parse['latex']

            self.expression_after_parser = to_parse['latex_before_cg']
            self.expression_after_grammar = parsed_expression
            self.parser_tree = parse.tree
            self.parser_list = parse.tlist
            self.lex_errors = to_parse['lex_errors']
            self.pure_lex_errors = to_parse['pure_lex_errors']
            self.yacc_errors = to_parse['yacc_errors']
            self.pure_yacc_errors = to_parse['pure_yacc_errors']

            return parsed_expression

        except Exception as e:
            print("[api.py] __to_parse | Exception:")
            raise e

    def recognize(self, image):
        try:
            if isinstance(image, str):
                image = cv.imread(image, 0)
            else:
                image = image.copy()

            self.image = image

            if self.image is not None:
                hme = Recognize(image)

                expression, image = hme.to_recognize()

                self.expression_after_recognition = expression.copy()
                self.predictions = hme.prediction

                parsed_expression = self.__to_parse(expression)

                self.parsed_expression = parsed_expression
                self.processed_image = image

                return parsed_expression, image
            else:
                print("[api.py] recognize | Exception:")
                raise Exception("You must enter an image.")
        except Exception as e:
            print("[api.py] recognize | Exception:")
            raise e

    def reset(self):
        self.image=None
        self.parsed_expression=""
        self.processed_image=None
        self.predictions=None
        self.configurations=None
        self.expression_after_recognition={} 
        self.expression_after_parser=[]
        self.expression_after_grammar=""
        self.parser_tree=None
        self.parser_list=None
        self.lex_errors=None
        self.yacc_errors=None
        self.pure_lex_errors=None
        self.pure_yacc_errors=None

    def get_predictions(self):
        return self.predictions

    def get_labels(self):
        labels = helpers.get_labels()
        return labels

    def get_expression_before_parser(self):
        return self.expression_before_parser

    def get_expression_after_parser(self):
        return self.expression_after_parser

    def get_expression_before_grammar(self):
        return self.expression_before_grammar

    def get_expression_after_grammar(self):
        return self.expression_after_grammar

    def get_lex_errors(self):
        return self.lex_errors

    def get_yacc_errors(self):
        return self.yacc_errors

    def get_pure_lex_errors(self):
        return self.pure_lex_errors

    def get_pure_yacc_errors(self):
        return self.pure_yacc_errors

    def configuration(self):
        return config



import numpy as np
from rhme import helpers
from rhme.hme_parser.grammar import lex as lex
from rhme.hme_parser.grammar import yacc as yacc

helpers_labels = helpers.get_labels()
labels = helpers_labels['labels_parser']

class BaseGrammar():

    def __init__(self):
        pass

    def correct_grammar_lex(self, token_errors, latex, latex_list, index=0, previous_attemptions=[]):
        """Fix grammar and lex errors.
        It tries to fix one error at the time. 
        The error index is passed as parameter, otherwise, the first error (index=0) will be used.

        Args:
            token_errors (list): Errors found in previous step: either check_grammar_lex.py or check_grammar_sintax.py
            latex (list): First latex structure.
            latex_list (list): [description]
            index (int, optional): Index of error. Default to 0.
            previous_attemptions (list, optional): This list stores all previous symbol (index) attemptions to fix the error. Defaults to [].

        Returns:
            latex_string, token_errors, index: Updated data.
        """
        latex_string = ""

        helpers.debug("[base_grammar.py] correct_grammar_lex()")
        helpers.debug("[base_grammar.py] correct_grammar_lex() | List of errors:")
        helpers.debug(token_errors)

        if len(token_errors) > 0:

            print('index: ', index)
            pos = token_errors[index]['pos']
            pos_list = token_errors[index]['pos_list']
            pred = token_errors[index]['prediction'].copy()
            print('pos: ', pos)

            helpers.debug("[base_grammar.py] correct_grammar_lex() | Error position")

            # colocar em um helpers
            subst = {
                'frac': {
                    '-': "\\frac",
                    'above': '',
                },
                'below': {
                    'below': ''
                },
                'sqrt': {
                    'sqrt': '\\sqrt'
                },
                'super': {
                    'super': '^'
                },
                'mult': {
                    '*': '\cdot'
                }
            }

            if not isinstance(pred, list):

                json_label = 'labels_parser'

                def get_new_index(pred):
                    helpers.debug("[base_grammar.py] correct_grammar_lex() | Reset prediction of current symbol")
                    new_pred = pred.copy()
                    new_pred[0][np.argmax(pred)] = 0
                    helpers.debug("[base_grammar.py] correct_grammar_lex() | Gets new indx and prediction from next index with higher prediction")
                    new_index = np.argmax(new_pred)
                    return new_index, new_pred

                def recur_get_new_index(pred):
                    new_index, pred = get_new_index(pred)
                    if new_index in token_errors[index]['attempts'] or new_index in previous_attemptions:
                        helpers.debug("[base_grammar.py] correct_grammar_lex() | New index is in previous attempts. Getting next.")
                        return recur_get_new_index(pred)
                    else:
                        label_recog = helpers_labels[json_label][str(new_index)]
                        new_label = helpers_labels["labels_recognition"][label_recog]
                        new_identification = labels[new_label]
                        helpers.debug("[base_grammar.py] correct_grammar_lex() | New symbol identification: %s " % new_identification)
                        return new_index, pred, new_identification

                new_index, new_pred, new_identification = recur_get_new_index(pred)

                helpers.debug("[base_grammar.py] correct_grammar_lex() | Updated symbol prediction")
                token_errors[index]['prediction'] = new_pred

                helpers.debug("[base_grammar.py] correct_grammar_lex() | Store new index attempt")
                token_errors[index]['attempts'].append(new_index)

                helpers.debug("[base_grammar.py] correct_grammar_lex() | Updated 'latex' and 'latex_list'. It is NOT a copy so, it is already updated.")

                # Make ir more 'Pythonic' later
                if new_identification in subst:
                    for substitution in subst[new_identification]:
                        if new_identification == substitution:
                            new_identification = subst[new_identification][substitution]

                latex_list[pos_list] = new_identification
                latex[pos_list]['label'] = new_identification
                latex[pos_list]['prediction'] = new_pred

                helpers.debug("[base_grammar.py] correct_grammar_lex() | Creating new 'latex_string' from 'latex_list'")
                latex_string = latex_string.join(latex_list)

                print("new: ", latex_string)
        return latex_string, token_errors, index